//-----------------------------------------------------------------------------
// Copyright (C) 1997, Paul McCarthy.
//
// MailCompactor.cc
//
//	This program compacts NeXT mail boxes, by removing mail messages
//	that have been flagged for deletion.
//
//	Tested with:
//		* NEXTSTEP 3.3
//		* OPENSTEP 4.1 (mach)
//		* OPENSTEP 4.2 (mach)
//
// NO WARRANTY:
//	USE THIS PROGRAM AT YOUR OWN RISK.  This program is provided
//	"as-is".  There is no warranty.  There is no claim of
//	merchantability nor fitness for any particular purpose.  The
//	authors will not accept liablity for any damage or loss directly
//	or indirectly attributed to the use of this program.
//
// FREEWARE:
//	You may use this program for free.  You may make and distribute
//	copies of this program provided that: (a) the program and these
//	notices are not modified in any way, and (b) no fee of any kind
//	is collected for the distribution.
//
// USAGE:
//	This program is generally used by site system administrators to
//	compact users' mailboxes for them on a nightly basis.  This should
//	be set up as a cron job.  Here's the command that we use at mcgh:
//
//	HOMES=/system2/Homes
//	COMPACTOR=/usr/local/bin/MailCompactor
//
//	find $HOMES -name '*.mbox' -type d -print \
//		-exec $COMPACTOR {} \; -prune
//
// ARCHITECTURE:
//	This program inspects the original table_of_contents to determine 
//	if there are any deleted messages.  If there are deleted messages 
//	the program will copy all undeleted messages from the mailbox and 
//	table_of_contents files into temporary files and record the names 
//	of the NeXT mail attachments of all deleted messages.  If (and 
//	only if) the entire mailbox is processed succesfully, then the 
//	temporary files are renamed to replace the originals, and the 
//	attachments from the deleted messages are removed.  
//
//	This architecture minimizes the opportunities for accidental 
//	mailbox corruption.  The original mailbox is left untouched until 
//	a completely finished and consistent new mailbox is ready.  The 
//	only opportunity for a corrupt mailbox occurs in the space of time 
//	between renaming the two files, and even in that case, the files 
//	will be self-consistent, there will only be an inconsistency 
//	between the mbox file and the table_of_contents file.  
//
// ACKNOWLEDGEMENTS:
//	Decoding of the NeXT mail table_of_contents is taken from the 
//	mailapp-utilities package written by Carl Edman, who got it from 
//	Chris Paris.  The lock routine and directory removal routines are 
//	also adaptations from that package.  
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MailCompactor.cc,v 1.2 97/12/15 20:23:57 zarnuk Exp $
// $Log:	MailCompactor.cc,v $
//  Revision 1.2  97/12/15  20:23:57  zarnuk
//  v2 -- Handled compiler warnings under OPENSTEP.
//  
//  Revision 1.1  97/12/15  14:59:10  zarnuk
//  v1 -- compact NeXT mail boxes.
//-----------------------------------------------------------------------------
extern "C" {
#ifdef __NeXT__
#include <libc.h>
#else
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>	// gethostname()
#include <sys/param.h>
#include <fcntl.h>
#endif
#include <assert.h>
#include <errno.h>
}

#ifndef HAVE_BOOL
#define	bool	int
#define	true	1
#define	false	0
#endif

#ifdef __LITTLE_ENDIAN__
static void swap_bytes( void* t )
    {
    char c; char* const p = (char*) t;
    c = p[0]; p[0] = p[3]; p[3] = c;
    c = p[1]; p[1] = p[2]; p[2] = c;
    }
#define	SWAP_BYTES(X)	swap_bytes( &(X) );
#else
#define	SWAP_BYTES(X)
#endif

int const PROG_VER = 2;
static char const PROG_NAME[] = "MailCompactor";
static char const TEMP_MBOX[] = "MailCompactor.box";
static char const TEMP_TOC[]  = "MailCompactor.toc";
static char const TOC_FILE[]  = "table_of_contents";
static char const LOCK_FILE[] = ".lock";
static char const MBOX_FILE[] = "mbox";
static char const* MBOX_DIR = 0;
static bool HAVE_LOCK = false;


//-----------------------------------------------------------------------------
// quit
//-----------------------------------------------------------------------------
#ifdef __GNUC__
static void quit( int rc ) __attribute__((noreturn));
#endif
static void quit( int rc )
    {
    remove( TEMP_TOC );
    remove( TEMP_MBOX );
    if (HAVE_LOCK)
	remove( LOCK_FILE );
    exit( rc );
    }


//-----------------------------------------------------------------------------
// error
//-----------------------------------------------------------------------------
#ifdef __GNUC__
static void error( int err, char const* fmt, ... )
	__attribute__((noreturn,format(printf,2,3)));
#endif
static void error( int err, char const* fmt, ... )
    {
    va_list args;
    fprintf( stderr, "%s: ", PROG_NAME );
    va_start( args, fmt );
    vfprintf( stderr, fmt, args );
    va_end( args );
    if (err != 0)
	fprintf( stderr, ": %s", strerror( err ) );
    fprintf( stderr, "\n" );
    fflush( stderr );
    quit( 2 );
    }


//=============================================================================
// SYSTEM CALLS
//-----------------------------------------------------------------------------

static void do_fclose( FILE* fp, char const* name )
    {
    if (fclose( fp ) != 0)
	error( errno, "fclose(%s/%s)", MBOX_DIR, name );
    }



static FILE* do_fopen( char const* file, char const* mode )
    {
    FILE* fp = fopen( file, mode );
    if (fp == 0)
	error( errno, "fopen(%s/%s)", MBOX_DIR, file );
    return fp;
    }



static void do_write( void const* buff, size_t sz, FILE* fp, char const* file )
    {
    if (fwrite( buff, sz, 1, fp ) != 1)
	error( errno, "fwrite(%s/%s)", MBOX_DIR, file );
    }



static void do_read( void* buff, size_t sz, FILE* fp, char const* file )
    {
    if (fread( buff, sz, 1, fp ) != 1)
	error( errno, "fread(%s/%s)", MBOX_DIR, file );
    }



static void do_fseek( FILE* fp, int pos, int whence, char const* file )
    {
    if (fseek( fp, pos, whence ) != 0)
	error( errno, "fseek(%s/%s,%d,%d)", MBOX_DIR, file, pos, whence );
    }



static void do_rename( char const* from, char const* to )
    {
    if (rename( from, to ) != 0)
	error( errno, "rename(%s/%s,%s/%s)", MBOX_DIR,from, MBOX_DIR,to );
    }



inline static void do_stat( char const* file, struct stat* st )
    {
    if (stat( file, st ) != 0)
	error( errno, "stat(%s/%s)", MBOX_DIR, file );
    }



static void do_remove( char const* file )
    {
    if (remove( file ) != 0 && errno != ENOENT)
	error( errno, "remove(%s/%s)", MBOX_DIR, file );
    }



static void* do_malloc( size_t n )
    {
    void* const p = malloc( n );
    if (p == 0)
	error( errno, "malloc(%lu)", n );
    return p;
    }



static void* do_realloc( void* t, size_t n )
    {
    void* const p = realloc( t, n );
    if (p == 0)
	error( errno, "realloc(%p,%lu)", t, n );
    return p;
    }



//-----------------------------------------------------------------------------
// skip_bytes
//-----------------------------------------------------------------------------
inline static void skip_bytes( size_t n, FILE* i, char const* s )
    { do_fseek( i, n, SEEK_CUR, s ); }


//-----------------------------------------------------------------------------
// copy_bytes
//-----------------------------------------------------------------------------
static void copy_bytes( size_t n, FILE* i, FILE* o,
			char const* s, char const* t )
    {
    char buff[ 512 ];
    while (n >= sizeof(buff))
	{
	do_read( buff, sizeof(buff), i, s );
	do_write( buff, sizeof(buff), o, t );
	n -= sizeof(buff);
	}
    if (n > 0)
	{
	do_read( buff, n, i, s );
	do_write( buff, n, o, t );
	}
    }


//-----------------------------------------------------------------------------
// match_stat
//-----------------------------------------------------------------------------
static void match_stat( char const* newfile, char const* oldfile, time_t* t )
    {
    struct stat st_old;
    struct stat st_new;

    do_stat( oldfile, &st_old );
    do_stat( newfile, &st_new );

    if (st_old.st_uid != st_new.st_uid || st_old.st_gid != st_new.st_gid)
	if (chown( newfile, st_old.st_uid, st_old.st_gid ) != 0)
	    error( errno, "chown(%s/%s)", MBOX_DIR, newfile );

    if (st_old.st_mode != st_new.st_mode)
	if (chmod( newfile, st_old.st_mode ) != 0)
	    error( errno, "chmod(%s/%s)", MBOX_DIR, newfile );

    *t = st_new.st_mtime;
    }


//=============================================================================
//-----------------------------------------------------------------------------
// TABLE_OF_CONTENTS
//
// NOTE *MSG_DATE*
//	The msg_date field is encoded as follows:
//	((year - 1900) << 9) | (month << 5) | day
//
// NOTE *MSG_STATUS*
//	NEXTSTEP 3.3:
//		read:' ','>' unread:'*' deleted:'d','D' flagged:'+'
//	Rhapsody 5.0 (Rhapsody Developer Release) on mach:
//		read:' ','>' unread:'*' deleted:'d','D' flagged:'+'
//
// NOTE *MSG_TYPE*
//	NEXTSTEP 3.3:
//		plain:' ' NeXT:'r' Mime:'m'.
//	Rhapsody 5.0 (Rhapsody Developer Release) on mach:
//		plain:' ' NeXT:'n' Mime:'m'.
//
// NOTE RHAPSODY:
//	The attach time and attach size fields are zeroed out for MIME
//	attachments under Rhapsody.  Also, most MIME messages under
//	Rhapsody create and delete the attachment on the fly.
//	Also the Mail.app SplitView and window size values are zeroed out
//	under Rhapsody.  They are kept in a separate "attrs.plist" file.
//	The "attrs.plist" file also contains some mailbox info that this
//	program is not maintaining yet, so this program cannot be used
//	for Rhapsody until those enhancements have been made.
//-----------------------------------------------------------------------------

int const TOC_MAGIC = 890712;

struct TOCHeader
	{
	int magic;		// File magic number.
	int num_msgs;		// Number of messages in mbox.
	time_t mbox_mtime;	// m_time of mbox.
	float height;		// Height of Mail.app SplitView.
	float frame_x;		// Mail.app window frame.
	float frame_y;
	float frame_w;
	float frame_h;
	};

struct TOCRec
	{
	size_t rec_len;	// Length of entry (including this word).
	size_t offset;	// Offset of this message in mbox.
	size_t length;	// Length in bytes of this message in mbox.
	int date;	// Date of message.  NOTE *MSG_DATE*
	char status;	// NOTE *MSG_STATUS*
	char type;	// NOTE *MSG_TYPE*
	char encrypted;	// Is this message encrypted?
	char pad;	// Always ' '.
	bool is_deleted() const	{ return status == 'd' || status == 'D'; }
	// followed by null-terminated "from" field.
	//	"(To: ...)" if from this user.
	// followed by null-terminated "subject" field.
	// followed by null-terminated "attachment" field.
	// The following fields are optional?
	// followed by attachment-size as a 4-byte big-endian int.
	// followed by attachment-mtime as a 4-byte big-endian int (time_t).
	// followed by priority as a 4-byte big-endian int.
	};


class TOCFile
	{
private:
	char const* toc_file;
	char const* mbox_file;
	FILE*	ftoc;
	TOCHeader hdr;
	TOCRec	rec;
	int	num_recs;
	int	rec_num;
	size_t	toc_size;
	size_t	toc_pos;
	size_t	mbox_size;
	size_t	mbox_pos;
	size_t	rec_len;
	size_t	dat_len;
	size_t	offset;
	size_t	length;
	size_t	buff_size;
	char*	buff;
	void expand_buffer( size_t );
public:
	TOCFile( char const* toc = TOC_FILE, char const* mbox = MBOX_FILE );
	~TOCFile();
	void close();
	void do_read( void* p, size_t n )
		{ ::do_read( p, n, ftoc, toc_file ); toc_pos += n; }
	void do_skip( size_t n )
		{ do_fseek( ftoc, n, SEEK_CUR, toc_file ); toc_pos += n; }
	bool read_rec();
	void read_data();
	void skip_data()		{ if (dat_len > 0) do_skip( dat_len ); }
	bool is_deleted() const		{ return rec.is_deleted(); }
	size_t get_length() const	{ return length; }
	void write_hdr( FILE* fp, char const* filename ) const
		{ do_write( &hdr, sizeof(hdr), fp, filename ); }
	void write_rec( size_t new_ofs, FILE* fp, char const* filename )
		{
		SWAP_BYTES( new_ofs );
		rec.offset = new_ofs;
		do_write( &rec, sizeof(rec), fp, filename );
		if (dat_len > 0)
		    do_write( buff, dat_len, fp, filename );
		}
	void patch_header( int nkeep, time_t mtime, char const* filename );
	char const* attachment() const;
	void check_excess() const;
	};


//-----------------------------------------------------------------------------
// expand_buffer
//-----------------------------------------------------------------------------
void TOCFile::expand_buffer( size_t goal )
    {
    while (buff_size < goal)
	buff_size += 2048;
    if (buff == 0)
	buff = (char*) do_malloc( buff_size );
    else
	buff = (char*) do_realloc( buff, buff_size );
    }


//-----------------------------------------------------------------------------
// close
//-----------------------------------------------------------------------------
void TOCFile::close()
    { if (ftoc != 0) do_fclose( ftoc, toc_file ); ftoc = 0; }


//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------
TOCFile::~TOCFile()
    {
    close();
    if (buff != 0) free( buff );
    }


//-----------------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------------
TOCFile::TOCFile( char const* toc, char const* mbox ):
    toc_file(toc), mbox_file(mbox), ftoc(0),
    rec_num(0), toc_size(0), toc_pos(0), mbox_size(0), mbox_pos(0),
    buff_size(0), buff(0)
    {
    struct stat st;

    do_stat( toc_file, &st );
    toc_size = st.st_size;

    do_stat( mbox_file, &st );
    mbox_size = st.st_size;

    if (toc_size < sizeof(hdr))
	error( 0, "%s/%s: File is too small (size=%lu min=%lu).",
		MBOX_DIR, toc_file, toc_size, sizeof(hdr) );

    ftoc = do_fopen( toc_file, "rb" );
    do_read( &hdr, sizeof(hdr) );

    int x = hdr.magic;	SWAP_BYTES( x );
    if (x != TOC_MAGIC)
	error( 0, "%s/%s: bad magic number %d, should be %d.",
		MBOX_DIR, toc_file, x, TOC_MAGIC );

    x = hdr.mbox_mtime;	SWAP_BYTES( x )
    if (x != st.st_mtime)
	error( 0, "%s/%s: not synchronized with %s.",
		MBOX_DIR, toc_file, mbox_file );

    x = hdr.num_msgs;	SWAP_BYTES( x )
    num_recs = x;
    }


//-----------------------------------------------------------------------------
// check_excess
//-----------------------------------------------------------------------------
void TOCFile::check_excess() const
    {
    if (toc_pos < toc_size)
	fprintf( stderr, "%s: %s/%s %lu extra bytes at end of file.\n",
		PROG_NAME, MBOX_DIR, TOC_FILE, (toc_size - toc_pos) );
    if (mbox_pos < mbox_size)
	fprintf( stderr, "%s: %s/%s %lu extra bytes at end of file.\n",
		PROG_NAME, MBOX_DIR, MBOX_FILE, (mbox_size - mbox_pos) );
    }


//-----------------------------------------------------------------------------
// patch_header
//-----------------------------------------------------------------------------
void TOCFile::patch_header( int nkeep, time_t mtime, char const* filename )
    {
    hdr.num_msgs = nkeep;	SWAP_BYTES( hdr.num_msgs )
    hdr.mbox_mtime = mtime;	SWAP_BYTES( hdr.mbox_mtime )
    FILE* fp = do_fopen( filename, "r+b" );
    do_fseek( fp, 0, SEEK_SET, filename );
    do_write( &hdr, sizeof(hdr), fp, filename );
    do_fclose( fp, filename );
    }


//-----------------------------------------------------------------------------
// read_rec
//-----------------------------------------------------------------------------
bool TOCFile::read_rec()
    {
    bool const have_one = (rec_num < num_recs);
    if (have_one)
	{
	if (toc_pos + sizeof(rec) > toc_size)
	    error( 0, "%s/%s truncated file.  %d/%d entries.",
		    MBOX_DIR, toc_file, rec_num, num_recs );
    
	do_read( &rec, sizeof(rec) );
    
	rec_num++;
	rec_len = rec.rec_len;	SWAP_BYTES( rec_len )
	dat_len = rec_len - sizeof(rec);
	offset  = rec.offset;	SWAP_BYTES( offset )
	length  = rec.length;	SWAP_BYTES( length )
    
	if (toc_pos + dat_len > toc_size ||
	    offset != mbox_pos || offset + length > mbox_size)
	    error( 0, "%s/%s corrupt entry: %d of %d, "
		    "size=%lu(max=%lu), offset=%lu(should be %lu) "
		    "length=%lu(max=%lu)",
		    MBOX_DIR, toc_file, rec_num, num_recs,
		    dat_len, toc_size - toc_pos, offset, mbox_pos,
		    length, mbox_size - mbox_pos );
    
	mbox_pos += length;
	}
    return have_one;
    }


//-----------------------------------------------------------------------------
// read_data
//-----------------------------------------------------------------------------
void TOCFile::read_data()
    {
    if (dat_len > 0)
	{
	if (dat_len > buff_size)
	    expand_buffer( dat_len );
	do_read( buff, dat_len );
	}
    }


//-----------------------------------------------------------------------------
// attachment
//-----------------------------------------------------------------------------
char const* TOCFile::attachment() const
    {
    char const* p = buff;
    char const* const plim = p + dat_len;
    bool ok = false;

    while (*p != 0 && p < plim)		// Skip "from" field
	p++;
    if (*p == 0 && p < plim)
	{
	p++;
	while (*p != 0 && p < plim)	// Skip "subject" field
	    p++;
	if (*p == 0 && p < plim)
	    {
	    p++;
	    ok = true;
	    }
	}

    if (!ok)
	error( 0, "%s/%s: record %d/%d, malformed string data.",
		MBOX_DIR, toc_file, rec_num, num_recs );

    return p;
    }


//=============================================================================
//-----------------------------------------------------------------------------
// PURGE_STACK
//-----------------------------------------------------------------------------
struct PurgeStack
	{
	PurgeStack* next;
	char s[1];		// room for null byte.
	};

//-----------------------------------------------------------------------------
// push
//-----------------------------------------------------------------------------
static PurgeStack* push( char const* s, PurgeStack* stk )
    {
    size_t const len = strlen(s);
    PurgeStack* const p = (PurgeStack*) do_malloc( sizeof(PurgeStack) + len );
    p->next = stk;
    memcpy( p->s, s, len + 1 );
    return p;
    }


//-----------------------------------------------------------------------------
// purge
//-----------------------------------------------------------------------------
static void purge( char const* file )
    {
    int const kid_pid = fork();

    if (kid_pid < 0)
	error( errno, "fork()" );

    else if (kid_pid == 0)
	{
	execl( "/bin/rm", "rm", "-rf", file, 0 );
	fprintf( stderr, "%s: execl(/bin/rm): %s\n",
		PROG_NAME, strerror( errno ) );
	_exit( 255 );
	}
    else
	{
	union wait w;
	int pid = -1;
	while ((pid = wait( &w )) >= 0 && pid != kid_pid) /* empty */;
	if (pid < 0)
	    error( errno, "wait()" );
	// FIXME: Check the exit status.
	}
    }


//-----------------------------------------------------------------------------
// purge
//-----------------------------------------------------------------------------
static void purge( PurgeStack* stk )
    {
    while (stk != 0)
	{
	PurgeStack* const p = stk;
	stk = stk->next;
	purge( p->s );
	free( p );
	}
    }


//=============================================================================
//-----------------------------------------------------------------------------
// do_compact
//-----------------------------------------------------------------------------
static void do_compact()
    {
    TOCFile toc;

    size_t opos = 0;
    size_t skip_len = 0;
    int nkeep = 0;
    int ntoss = 0;

    FILE* ibox = do_fopen( MBOX_FILE, "rb" );
    FILE* obox = do_fopen( TEMP_MBOX, "wb" );
    FILE* otoc = do_fopen( TEMP_TOC,  "wb" );
    PurgeStack* stk = 0;

    toc.write_hdr( otoc, TEMP_TOC );

    while (toc.read_rec())
	{
	toc.read_data();
	size_t const msg_len = toc.get_length();
	if (!toc.is_deleted())
	    {
	    nkeep++;
	    if (skip_len != 0)
		{
		skip_bytes( skip_len, ibox, MBOX_FILE );
		skip_len = 0;
		}
	    copy_bytes( msg_len, ibox, obox, MBOX_FILE, TEMP_MBOX );
	    toc.write_rec( opos, otoc, TEMP_TOC );
	    opos += msg_len;
	    }
	else
	    {
	    ntoss++;
	    skip_len += msg_len;
	    char const* const s = toc.attachment();
	    if (s != 0 && *s != 0)
		stk = push( s, stk );
	    }
	}
    toc.check_excess();

    do_fclose( otoc, TEMP_TOC );
    do_fclose( obox, TEMP_MBOX );
    do_fclose( ibox, MBOX_FILE );
    toc.close();

    if (ntoss != 0)
	{
	time_t mtime;
	match_stat( TEMP_MBOX, MBOX_FILE, &mtime );
	toc.patch_header( nkeep, mtime, TEMP_TOC );
	match_stat( TEMP_TOC, TOC_FILE, &mtime );
	do_rename( TEMP_MBOX, MBOX_FILE );
	do_rename( TEMP_TOC, TOC_FILE );
	purge( stk );
	}
    }


//-----------------------------------------------------------------------------
// need_compact
//-----------------------------------------------------------------------------
static bool need_compact()
    {
    TOCFile toc;

    while (toc.read_rec())
	{
	toc.skip_data();
	if (toc.is_deleted())
	    return true;
	}
    toc.check_excess();

    return false;
    }


//=============================================================================
//-----------------------------------------------------------------------------
// unlock_mbox
//-----------------------------------------------------------------------------
static void unlock_mbox()
    { HAVE_LOCK = false; do_remove( LOCK_FILE ); }


//-----------------------------------------------------------------------------
// lock_mbox
//-----------------------------------------------------------------------------
static bool lock_mbox()
    {
    char buff[ 1024 ];
    char myhost[ MAXHOSTNAMELEN ];

    gethostname( myhost, sizeof(myhost) );

    HAVE_LOCK = false;

    int fd = -1;
    while ((fd = open( LOCK_FILE, O_CREAT|O_EXCL|O_WRONLY, 0666 )) < 0)
	{
	if (errno != EEXIST)
	    error( errno, "create(%s/%s)", MBOX_DIR, LOCK_FILE );

	if ((fd = open( LOCK_FILE, O_RDONLY )) < 0)
	    error( errno, "open(%s/%s)", MBOX_DIR, LOCK_FILE );

	int n;
	if ((n = read( fd, buff, sizeof(buff)-1 )) < 0)
	    error( errno, "read(%s/%s)", MBOX_DIR, LOCK_FILE );
	buff[n] = '\0';

	close( fd );
	fd = -1;

	char* const host = buff;
	char* user = host;
	int ipid = -1;

	while (*user != 0 && *user != ',')
	    user++;

	if (*user != 0)
	    {
	    *user++ = '\0';
	    char* spid = user;
	    while (*spid != 0 && *spid != ',')
		spid++;
	    if (*spid != 0)
		{
		*spid++ = '\0';
		ipid = atoi( spid );
		}
	    }

	if (strcmp( host, myhost ) == 0 &&		// Stale lock?
	    ipid > 0 && kill( ipid, 0 ) != 0 && errno == ESRCH)
	    {
	    do_remove( LOCK_FILE );
	    }
	else
	    {
	    fprintf( stderr, "%s: %s locked by %s on %s (pid=%d)\n",
		PROG_NAME, MBOX_DIR, user, host, ipid );
	    break;
	    }
	}

    if (fd >= 0)
	{
	HAVE_LOCK = true;
	sprintf( buff, "%s,%s,%d", myhost, PROG_NAME, getpid() );
	int const n = strlen( buff );
	if (write( fd, buff, n ) != n)
	    error( errno, "write(%s/%s)", MBOX_DIR, LOCK_FILE );
	if (close( fd ) != 0)
	    error( errno, "close(%s/%s)", MBOX_DIR, LOCK_FILE );
	}

    return HAVE_LOCK;
    }


//-----------------------------------------------------------------------------
// usage
//-----------------------------------------------------------------------------
static void usage()
    {
    printf(
	"%s v%d\n"
	"\n"
	"This program compacts NeXT mail boxes.\n"
	"\n"
	"usage: %s dirname\n"
	"\n"
	"where: dirname is the mailbox directory to compact.\n"
	"\n", PROG_NAME, PROG_VER, PROG_NAME );
    }


//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main( int argc, char const* const* argv )
    {
    int rc = 1;
    MBOX_DIR = argv[1];

    if (argc != 2 ||
	strcmp( MBOX_DIR, "-h" ) == 0 ||
	strcmp( MBOX_DIR, "-help" ) == 0)
	usage();

    else if (chdir( MBOX_DIR ) != 0)
	fprintf( stderr, "%s: chdir(%s): %s\n",
		PROG_NAME, MBOX_DIR, strerror(errno) );

    else if (lock_mbox())
	{
	if (need_compact())
	    do_compact();
	unlock_mbox();
	rc = 0;
	}

    return rc;
    }
